package com.simutronics.eclipsefe;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionRegistry;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;

import com.simutronics.eclipsefe.extensions.ICommandProcessor;
import com.simutronics.eclipsefe.extensions.IInputProcessor;
import com.simutronics.eclipsefe.extensions.IMacroProcessor;
import com.simutronics.eclipsefe.extensions.IServiceProvider;
import com.simutronics.eclipsefe.preferences.FEPreferencePage;
import com.simutronics.eclipsefe.preferences.HighlightsPreferencePage;
import com.simutronics.eclipsefe.preferences.MacroPreferencePage;
import com.simutronics.eclipsefe.preferences.StoredInfoPreferencePage;
import com.simutronics.eclipsefe.preferences.VariablesPreferencePage;

/**
 * The main plugin class to be used in the desktop.
 */
public class EclipsefePlugin extends AbstractUIPlugin {
	//The shared instance.
	private static EclipsefePlugin plugin;
	//Resource bundle.
	private ResourceBundle resourceBundle;
	public static Bundle pluginBundle;
	
	private static final Map<String, IMacroProcessor> macroProcessors = new HashMap<String, IMacroProcessor>();
	private static final List<IInputProcessor> inputProcessors = new ArrayList<IInputProcessor>();
	private static final List<ICommandProcessor> commandProcessors = new ArrayList<ICommandProcessor>();
	
	private static final String MACRO_NAME = "string";
	private static final String MACRO_CHILD = "macro";
	private static final String MACRO_CLASS = "class";
	private static final String INPUT_PROCESSOR_CLASS = "class";
	private static final String COMMAND_PROCESSOR_CLASS = "class";
	private static final String SERVICE_PROVIDER_CLASS = "class";
	
	public static final String MACRO_EXTENSION_POINT = "com.simutronics.eclipsefe.macroprocessors";
	public static final String INPUT_PROCESSOR_EXTENSION_POINT = "com.simutronics.eclipsefe.inputprocessors";
	public static final String COMMAND_PROCESSOR_EXTENSION_POINT = "com.simutronics.eclipsefe.commandprocessors";
	public static final String SERVICE_PROVIDER_EXTENSION_POINT = "com.simutronics.eclipsefe.serviceproviders";
	
	public static final boolean DEBUG = false;
	
	public EclipsefePlugin() {
		super();
		plugin = this;
	}

	public void start(BundleContext context) throws Exception {
		super.start(context);
		pluginBundle = Platform.getBundle("com.simutronics.eclipsefe");
		setDefaultPreferences();
		buildProcessors();
	}
	
	private void setDefaultPreferences() {
		IPreferenceStore prefs = getPreferenceStore();
		FEPreferencePage.setDefaults(prefs);
		StoredInfoPreferencePage.setDefaults(prefs);
		MacroPreferencePage.setDefaults(prefs);
		HighlightsPreferencePage.setDefaults(prefs);
		VariablesPreferencePage.setDefaults(prefs);
	}
	
	private void buildProcessors() {
		IExtensionRegistry registry = Platform.getExtensionRegistry();
		IConfigurationElement[] macroEls = registry.getExtensionPoint(MACRO_EXTENSION_POINT).getConfigurationElements();
		for (IConfigurationElement el : macroEls) {
			try {
				IMacroProcessor macro = (IMacroProcessor) el.createExecutableExtension(MACRO_CLASS);
				IConfigurationElement[] children = el.getChildren(MACRO_CHILD);
				for (IConfigurationElement child : children) {
					String text = child.getAttribute(MACRO_NAME);
					if ((text != null) && text.matches("\\{.*\\}")) {
						macroProcessors.put(text, macro);
					}
				}
			} catch (CoreException e) {
				// go to the next element
			}
		}
		
		IConfigurationElement[] inputProcessorEls = registry.getExtensionPoint(INPUT_PROCESSOR_EXTENSION_POINT).getConfigurationElements();
		for (IConfigurationElement el : inputProcessorEls) {
			try {
				IInputProcessor processor = (IInputProcessor) el.createExecutableExtension(INPUT_PROCESSOR_CLASS);
				inputProcessors.add(processor);
			} catch (CoreException e) {
				// go to the next element
			}
		}
		
		IConfigurationElement[] commandProcessorEls = registry.getExtensionPoint(COMMAND_PROCESSOR_EXTENSION_POINT).getConfigurationElements();
		for (IConfigurationElement el : commandProcessorEls) {
			try {
				ICommandProcessor processor = (ICommandProcessor) el.createExecutableExtension(COMMAND_PROCESSOR_CLASS);
				commandProcessors.add(processor);
			} catch (CoreException e) {
				// go to the next element
			}
		}
	}

	public void stop(BundleContext context) throws Exception {
		super.stop(context);
		plugin = null;
		resourceBundle = null;
	}

	public static EclipsefePlugin getDefault() {
		return plugin;
	}

	public static String getResourceString(String key) {
		ResourceBundle bundle = EclipsefePlugin.getDefault().getResourceBundle();
		try {
			if (bundle != null) {
				return bundle.getString(key);
			}
			return key;
		} catch (MissingResourceException e) {
			return key;
		}
	}

	public ResourceBundle getResourceBundle() {
		try {
			if (resourceBundle == null) {
				resourceBundle = ResourceBundle.getBundle("com.simutronics.eclipsefe.EclipsefePluginResources");
			}
		} catch (MissingResourceException x) {
			resourceBundle = null;
		}
		return resourceBundle;
	}

	public static List<ICommandProcessor> getCommandProcessors() {
		return commandProcessors;
	}

	public static List<IInputProcessor> getInputProcessors() {
		return inputProcessors;
	}

	public static Map<String, IMacroProcessor> getMacroProcessors() {
		return macroProcessors;
	}
	
	public static List<IServiceProvider> getServiceProviders() {
		ArrayList<IServiceProvider> providers = new ArrayList<IServiceProvider>();
		IExtensionRegistry registry = Platform.getExtensionRegistry();

		IConfigurationElement[] serviceProviderEls = registry.getExtensionPoint(SERVICE_PROVIDER_EXTENSION_POINT).getConfigurationElements();
		for (IConfigurationElement el : serviceProviderEls) {
			try {
				IServiceProvider provider = (IServiceProvider) el.createExecutableExtension(SERVICE_PROVIDER_CLASS);
				providers.add(provider);
			} catch (CoreException e) {
				e.getMessage();
				// go to the next element
			}
		}
		return providers;
	}
}
