/*
 * Created on Nov 5, 2005
 */
package com.simutronics.eclipsefe.scripting;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StreamTokenizer;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.simutronics.eclipsefe.FEView;
import com.simutronics.eclipsefe.VariableService;

/**
 * @author Adam Vartanian
 */
public class ScriptThread extends Thread {

	private static class Match {
		public String value = null;
		public Pattern regex = null;
		public int line;
		
		public Match(String s, int l) {
			value = s;
			line = l;
		}
		
		public Match(Pattern p, int l) {
			regex = p;
			line = l;
		}
	}
	
	private static class ScriptException extends Exception {
		public ScriptException(String msg) {
			super(msg);
		}
	}
	
	private FEView window;
	private ArrayList<String[]> lines = new ArrayList<String[]>();
	private Map<String, Integer> labels = new HashMap<String, Integer>();
	private int current = 1;
	private boolean waitingForRoomDesc = false;
	private boolean waitingForMatch = false;
	private boolean waiting = false;
	private boolean waitingForPrompt = false;
	private ArrayList<Match> matchtable = new ArrayList<Match>();
	private String waitingForText = null;
	private Pattern waitingForRE = null;
	
	public ScriptThread(FEView window, File script, String args) throws IOException {
		super();
		this.window = window;
		BufferedReader isr = new BufferedReader(new InputStreamReader(new FileInputStream(script)));
		int i = 1;
		lines.add(null);   // blank line 0
		String line = isr.readLine();
		while (line != null) {
			line = line.trim();
			if ((line.length() > 0) && (line.charAt(0) == '#')) {
				line = "";
			}
			if ((line.length() > 0) && (line.charAt(line.length() - 1) == ':')) {
				labels.put(line.substring(0, line.length() - 1), i);
				line = "";
			}
			if (line.length() > 0) {
				String[] segments = line.split(" +", 2);
				if (segments.length == 2) {
					lines.add(segments);
				} else {
					lines.add(new String[]{segments[0], ""});
				}
			} else {
				lines.add(null);
			}
			i++;
			line = isr.readLine();
		}
		window.getVariableService().setVariable("0", args.trim());
		tokenizeArgs();
		matchtable.clear();
	}
	
	public void run() {
		VariableService varServ = null;
		try {
			while (true) {
				String[] line = lines.get(current);
				if (line != null) {
					varServ = window.getVariableService();
					executeCommand(varServ.variableReplacement(line[0]).toLowerCase(), varServ.variableReplacement(line[1]), varServ);
				}
				current++;
				if (current >= lines.size()) {
					return;
				}
			}
		} catch (ScriptException e) {
			window.displayError(e.getMessage());
		} finally {
			varServ = window.getVariableService();
			varServ.deleteVariable("0");
			varServ.deleteVariable("1");
			varServ.deleteVariable("2");
			varServ.deleteVariable("3");
			varServ.deleteVariable("4");
			varServ.deleteVariable("5");
			varServ.deleteVariable("6");
			varServ.deleteVariable("7");
			varServ.deleteVariable("8");
			varServ.deleteVariable("9");
		}
	}
	
	private void executeCommand(String command, String remaining, VariableService varServ) throws ScriptException {
		if (command.equals("counter")) {
			String[] segments = remaining.split(" +", 2);
			segments[0] = segments[0].toLowerCase();
			int arg = 0;
			try {
				if (segments.length != 2) {
					throw new ScriptException("Not enough arguments for counter on line " + current);
				}
				arg = Integer.parseInt(segments[1]);
			} catch (NumberFormatException e) {
				throw new ScriptException("Non-numeric argument [" + segments[1] + " given to counter on line " + current);
			}
			try {
				if (segments[0].equals("set")) {
					varServ.setCounter(arg);
				} else if (segments[0].equals("add")) {
					varServ.setCounter(varServ.getCounter() + arg);
				} else if (segments[0].equals("subtract")) {
					varServ.setCounter(varServ.getCounter() - arg);
				} else if (segments[0].equals("multiply")) {
					varServ.setCounter(varServ.getCounter() * arg);
				} else if (segments[0].equals("divide")) {
					varServ.setCounter(varServ.getCounter() / arg);
				} else {
					throw new ScriptException("Invalid counter command [" + segments[0] + "] on line " + current);
				}
			} catch (NumberFormatException e) {
				throw new ScriptException("Invalid counter value.");
			}
		} else if (command.equals("deletevariable")) {
			if (!varServ.deleteVariable(remaining)) {
				throw new ScriptException("Attempt to delete nonexistant variable [" + remaining + "] on line " + current);
			}
		} else if (command.equals("echo")) {
			window.appendText(remaining + '\n');
		} else if (command.equals("echow")) {
			String[] segments = remaining.split(" +", 2);
			window.appendText(segments[0], segments[1] + '\n', true);
		} else if (command.equals("exit")) {
			return;
		} else if (command.equals("goto")) {
			if (!labels.containsKey(remaining)) {
				throw new ScriptException("Could not find label [" + remaining + "] in goto on line " + current);
			}
			current = labels.get(remaining);
		} else if (command.startsWith("if_")) {
			String var = command.substring(3);
			if (var.length() == 0) {
				throw new ScriptException("Invalid script command [" + command + "] found at line " + current);
			}
			String value = varServ.getVariable(var);
			if ((value != null) && (value.length() > 0)) {
				String[] segments = remaining.split(" +", 2);
				if (segments.length == 1) {
					executeCommand(segments[0].toLowerCase(), "", varServ);
				} else {
					executeCommand(segments[0].toLowerCase(), segments[1], varServ);
				}
			}
		} else if (command.equals("match")) {
			String[] segments = remaining.split(" +", 2);
			if (segments.length != 2) {
				throw new ScriptException("No match string on line " + current);
			}
			if (!labels.containsKey(segments[0])) {
				throw new ScriptException("Could not find label [" + segments[0] + "] in match on line " + current);
			}
			matchtable.add(new Match(segments[1], labels.get(segments[0])));
		} else if (command.equals("matchre")) {
			String[] segments = remaining.split(" +", 2);
			if (segments.length != 2) {
				throw new ScriptException("No match string on line " + current);
			}
			if (!labels.containsKey(segments[0])) {
				throw new ScriptException("Could not find label [" + segments[0] + "] in matchre on line " + current);
			}
			matchtable.add(new Match(Pattern.compile(segments[1]), labels.get(segments[0])));
		} else if (command.equals("matchwait")) {
			waitingForMatch = true;
			doWait();
			matchtable.clear();
		} else if (command.equals("move")) {
			window.sendCommand(remaining, false);
			nextroom();
		} else if (command.equals("nextroom")) {
			nextroom();
		} else if (command.equals("pause")) {
			try {
				int time;
				if (remaining.length() == 0) {
					time = 1;
				} else {
					time = Integer.parseInt(remaining);
				}
				time *= 1000;
				sleep(time);
			} catch (NumberFormatException e) {
				throw new ScriptException("Could not parse a number from [" + remaining + "] on line " + current);
			} catch (InterruptedException e) {
				// continue
			}
		} else if (command.equals("put")) {
			window.sendCommand(remaining, false);
		} else if (command.equals("save")) {
			varServ.setVariable("s", remaining);
		} else if (command.equals("set")) {
			setvariable(remaining, varServ);
		} else if (command.equals("setvariable")) {
			setvariable(remaining, varServ);
		} else if (command.equals("shift")) {
			String args = varServ.getVariable("0");
			if (args.length() > 0) {
				if (args.charAt(0) == '\"') {
					args = args.substring(args.indexOf('\"', 1) + 1);
				} else {
					args = args.substring(args.indexOf(' '));
				}
				varServ.setVariable("0", args.trim());
				tokenizeArgs();	
			}
		} else if (command.equals("wait")) {
			waitingForPrompt = true;
			doWait();
		} else if (command.equals("waitfor")) {
			if (remaining.length() > 0) {
				waitingForText = remaining;
				doWait();
			}
		} else if (command.equals("waitforre")) {
			if (remaining.length() > 0) {
				waitingForRE = Pattern.compile(remaining);
				doWait();
			}
		} else {
			throw new ScriptException("Invalid script command [" + command + "] found at line " + current);
		}

	}
	
	public void processRawString(String string) {
		if (waitingForRoomDesc) {
			if (string.indexOf("<style id=\"roomName\" />") > -1) {
				stopWaiting();
			}
		} else if (waitingForPrompt) {
			if (string.indexOf("<prompt ") > -1) {
				stopWaiting();
			}
		}
	}
	
	public void processParsedString(String string) {
		if (waitingForText != null) {
			if (string.indexOf(waitingForText) > -1) {
				stopWaiting();
			}
		} else if (waitingForRE != null) {
			Matcher m = waitingForRE.matcher(string);
			if (m.find()) {
				stopWaiting();
			}
		} else if (waitingForMatch) {
			if (match(string)) {
				stopWaiting();
			}
		}
	}
	
	private void nextroom() {
		waitingForRoomDesc = true;
		doWait();
	}
	
	private void setvariable(String remaining, VariableService varServ) {
		String[] segments = remaining.split(" +", 2);
		if (segments.length == 1) {
			varServ.setVariable(segments[0], "");
		} else {
			varServ.setVariable(segments[0], segments[1]);
		}
	}
	
	private void doWait() {
		waiting = true;
		while (waiting) {
			try {
				sleep(5000);
			} catch (InterruptedException e) {
				// go back to sleeping
			}
		}
	}
	
	private void stopWaiting() {
		waitingForRoomDesc = false;
		waitingForPrompt = false;
		waitingForMatch = false;
		waitingForText = null;
		waitingForRE = null;
		waiting = false;
		interrupt();
		Thread.yield();
	}
	
	private void tokenizeArgs() {
		VariableService varServ = window.getVariableService();
		String args = varServ.getVariable("0");
		StreamTokenizer st = new StreamTokenizer(new StringReader(args));
		st.quoteChar('\"');
		st.ordinaryChars('0', '9');
		st.ordinaryChar('-');
		st.ordinaryChar('.');
		st.wordChars('_', '_');
		st.wordChars('-', '-');
		st.wordChars('0', '9');
		st.wordChars('.', '.');
		int i = 1;
		try {
			while ((st.nextToken() != StreamTokenizer.TT_EOF) && (i <= 9)) {
				if (st.ttype == StreamTokenizer.TT_WORD) {
					varServ.setVariable(Integer.toString(i), st.sval.replace('_', ' '));
				} else if (st.ttype == '\"') {
					varServ.setVariable(Integer.toString(i), st.sval);
				}
				i++;
			}
			while (i <= 9) {
				varServ.setVariable(Integer.toString(i), "");
				i++;
			}
		} catch (IOException e) {
			// An IO exception from a StringReader?  WTF?
		}
	}
	
	private boolean match(String s) {
		for (Match m : matchtable) {
			if ((m.value != null) && (s.indexOf(m.value) > -1)) {
				current = m.line;
				return true;
			} else if (m.regex != null) {
				Matcher matcher = m.regex.matcher(s);
				if (matcher.find()) {
					current = m.line;
					return true;
				}
			}
		}
		return false;
	}

}
